#![feature(round_char_boundary)]
#![feature(exact_size_is_empty)]
use std::io::Read;
use std::path::Path;

use ::organic::parser::parse;
use organic::parser::parse_with_settings;
use organic::settings::GlobalSettings;
use organic::settings::LocalFileAccessInterface;

#[cfg(feature = "tracing")]
use crate::init_tracing::init_telemetry;
#[cfg(feature = "tracing")]
use crate::init_tracing::shutdown_telemetry;
#[cfg(feature = "tracing")]
mod init_tracing;

#[cfg(not(feature = "tracing"))]
fn main() -> Result<(), Box<dyn std::error::Error>> {
    main_body()
}

#[cfg(feature = "tracing")]
fn main() -> Result<(), Box<dyn std::error::Error>> {
    let rt = tokio::runtime::Runtime::new()?;
    let result = rt.block_on(async {
        init_telemetry()?;
        let main_body_result = main_body();
        shutdown_telemetry()?;
        main_body_result
    });
    result
}

#[cfg_attr(feature = "tracing", tracing::instrument(ret, level = "debug"))]
fn main_body() -> Result<(), Box<dyn std::error::Error>> {
    let args = std::env::args().skip(1);
    if args.is_empty() {
        let org_contents = read_stdin_to_string()?;
        run_anonymous_parse(org_contents)
    } else {
        for arg in args {
            run_parse_on_file(arg)?
        }
        Ok(())
    }
}

fn read_stdin_to_string() -> Result<String, Box<dyn std::error::Error>> {
    let mut stdin_contents = String::new();
    std::io::stdin()
        .lock()
        .read_to_string(&mut stdin_contents)?;
    Ok(stdin_contents)
}

fn run_anonymous_parse<P: AsRef<str>>(org_contents: P) -> Result<(), Box<dyn std::error::Error>> {
    let rust_parsed = parse(org_contents.as_ref())?;
    println!("{:#?}", rust_parsed);
    Ok(())
}

fn run_parse_on_file<P: AsRef<Path>>(org_path: P) -> Result<(), Box<dyn std::error::Error>> {
    let org_path = org_path.as_ref();
    let parent_directory = org_path
        .parent()
        .ok_or("Should be contained inside a directory.")?;
    let org_contents = std::fs::read_to_string(org_path)?;
    let org_contents = org_contents.as_str();
    let file_access_interface = LocalFileAccessInterface {
        working_directory: Some(parent_directory.to_path_buf()),
    };
    let global_settings = {
        let mut global_settings = GlobalSettings::default();
        global_settings.file_access = &file_access_interface;
        global_settings
    };
    let rust_parsed = parse_with_settings(org_contents, &global_settings)?;
    println!("{:#?}", rust_parsed);
    Ok(())
}
