use super::element::Element;
use super::lesser_element::TableCell;
use super::Keyword;
use super::Object;
use super::StandardProperties;

#[derive(Debug)]
pub struct PlainList<'s> {
    pub source: &'s str,
    pub list_type: PlainListType,
    pub children: Vec<PlainListItem<'s>>,
}

#[derive(Debug, Copy, Clone)]
pub enum PlainListType {
    Unordered,
    Ordered,
    Descriptive,
}

/// The width that something is indented. For example, a single tab character could be a value of 4 or 8.
pub type IndentationLevel = u16;

#[derive(Debug)]
pub struct PlainListItem<'s> {
    pub source: &'s str,
    pub indentation: IndentationLevel,
    pub bullet: &'s str,
    pub counter: Option<PlainListItemCounter>,
    pub checkbox: Option<(CheckboxType, &'s str)>,
    pub tag: Vec<Object<'s>>,
    pub pre_blank: PlainListItemPreBlank,
    pub children: Vec<Element<'s>>,
}

pub type PlainListItemCounter = u16;
pub type PlainListItemPreBlank = u8;

#[derive(Debug)]
pub enum CheckboxType {
    On,
    Trans,
    Off,
}

#[derive(Debug)]
pub struct GreaterBlock<'s> {
    pub source: &'s str,
    pub name: &'s str,
    pub parameters: Option<&'s str>,
    pub children: Vec<Element<'s>>,
}

#[derive(Debug)]
pub struct DynamicBlock<'s> {
    pub source: &'s str,
    pub name: &'s str,
    pub parameters: Option<&'s str>,
    pub children: Vec<Element<'s>>,
}

#[derive(Debug)]
pub struct FootnoteDefinition<'s> {
    pub source: &'s str,
    pub label: &'s str,
    pub children: Vec<Element<'s>>,
}

#[derive(Debug)]
pub struct Drawer<'s> {
    pub source: &'s str,
    pub name: &'s str,
    pub children: Vec<Element<'s>>,
}

#[derive(Debug)]
pub struct PropertyDrawer<'s> {
    pub source: &'s str,
    pub children: Vec<NodeProperty<'s>>,
}

#[derive(Debug)]
pub struct NodeProperty<'s> {
    pub source: &'s str,
    pub value: Option<&'s str>,
}

#[derive(Debug)]
pub struct Table<'s> {
    pub source: &'s str,
    pub formulas: Vec<Keyword<'s>>,
    pub children: Vec<TableRow<'s>>,
}

#[derive(Debug)]
pub struct TableRow<'s> {
    pub source: &'s str,
    pub children: Vec<TableCell<'s>>,
}

impl<'s> StandardProperties<'s> for PlainList<'s> {
    fn get_source(&'s self) -> &'s str {
        self.source
    }
}

impl<'s> StandardProperties<'s> for PlainListItem<'s> {
    fn get_source(&'s self) -> &'s str {
        self.source
    }
}

impl<'s> StandardProperties<'s> for GreaterBlock<'s> {
    fn get_source(&'s self) -> &'s str {
        self.source
    }
}

impl<'s> StandardProperties<'s> for DynamicBlock<'s> {
    fn get_source(&'s self) -> &'s str {
        self.source
    }
}

impl<'s> StandardProperties<'s> for FootnoteDefinition<'s> {
    fn get_source(&'s self) -> &'s str {
        self.source
    }
}

impl<'s> StandardProperties<'s> for Drawer<'s> {
    fn get_source(&'s self) -> &'s str {
        self.source
    }
}

impl<'s> StandardProperties<'s> for PropertyDrawer<'s> {
    fn get_source(&'s self) -> &'s str {
        self.source
    }
}

impl<'s> StandardProperties<'s> for NodeProperty<'s> {
    fn get_source(&'s self) -> &'s str {
        self.source
    }
}

impl<'s> StandardProperties<'s> for Table<'s> {
    fn get_source(&'s self) -> &'s str {
        self.source
    }
}

impl<'s> StandardProperties<'s> for TableRow<'s> {
    fn get_source(&'s self) -> &'s str {
        self.source
    }
}

impl<'s> PlainListItem<'s> {
    pub fn get_checkbox(&self) -> Option<&'s str> {
        self.checkbox.as_ref().map(|(_, checkbox)| *checkbox)
    }

    pub fn get_checkbox_type(&self) -> Option<&CheckboxType> {
        self.checkbox
            .as_ref()
            .map(|(checkbox_type, _)| checkbox_type)
    }
}
